///////////////////////////////////////////////////////////////////////////////
//
//				LESZIFR3 (STRONG)
//				   (software)
//
//				   for Linux
//				 
//				  ver.0.4 (c)
//				   
//   Plik szyfrowany jest w szumie. Zastosowane sa dwa tryby szumu:
//   1. Gesty - da plikow o rozmiarze bitowym ponizej limitu 
//      SPEED_CODING_LIMIT (defaut - dla rozmiau pliku ponizej 3MB.
//      Mozna korzystac z dwoch typow Gestego generatora:
//      - generator liniowy (prandom_gen.c) (*)
//      - generator nieliniowy (prandom_nonlinear.c) (*)
//   2. Szybki dla plikow o objetosci bitowej powyzej (lub rownej)
//      SPEED_CODING_LIMIT (default - powyzej 3MB)
//   
//   (*) wybor generatora liniowy/nieliniowy jest w sekcji #include 
//       w pliku ngen1.c (wymaga ponownej kompilacji leszifr3)
//
//   Uwaga! wersja leszif3 (b) nie jest kompatibilna z leszif3 (a)
//
//   Generator szumu jest generatorem 1/2-pseudolosowm. Entropia jest
//   pobierana z dwoch zrodel dla trybu Gestego:
//   1. Pierwsze zrodlo towrzy SEEDs z generatora sprzetowego RDRAND (CPU)
//   2. Drugim zrdodlem jest niemotoniczna funkcja czasu dzialajaca
//      na ziarnach nanosekundowych REAL CLOCK.
//   W trybie Gestego tworzenia szumu binarnego SEEDs sa kombinacja
//   obydwu zrodel dla dwoch kanalow tworzenia szumu. 
//   Nalezy zwrocic uwage, ze starsze CPU (Intel i3/i5/i7) drugiej
//   gneracji nie posiadaja RDRAND. Oprogramowanie rozpoznaje CPU,
//   i jesli to nie jest mozliwe pomija generowanie SEEDs z pomoca
//   RDRAND i SEEDs pochodza z niemonotonicznej funkcji czasu. 
//   Okres Gestego generatora szumu wynosi 2^62.
//   Szybki (Fast) generator szumu binarnego ma dwa losowe kanaly
//   tworzenia szumu przy SEEDs generowanym przez niemotoniczna 
//   funckje czasu. Okres generatora wynosi 2^48.
//
//   Hasla sa szyfrowane wstepnie funckja skrotu (hash) na zmodyfikowanej MD5.
//   Nastepnie wyliczone hashe sa uzyte jako SEEDs dla gneratora entropi
//   RSA. 
//   Szum binarny jest wyliczany dla kazdego bloku danych (max 4kB). 
//   W sumie szum binarny ma dlugosc rowna dlugosci pliku. 
//   Kodowanie przebiega nastepujaco dla kazdego bloku:
//   A - tekst (lub binarka) przeznaczone do szyfrowania
//   B - szum binarny (z generatorow Gestego lub Fast. 
//   C - Szyfrowanie klucza za pomoca generatora RSA.
//   Krok 0:
//        C = pasword szyfrowany generatorem RSA (blok 4k)
//   Krok 1: 
//        B1 = A xor B
//   Krok 2:
//        D = C xor B
//   Krok 3:
//        Result = (permutacja bitowa -> B1)  xor D
//   do pliku szyfrowanego zapisywana jest suma prosta B1 oraz Result tj.
//        blok szyfrowany = B1 (+) Result
//   oraz dodawana jest zmodyfikowana MD5  bloku przed
//   przed szyfrowaniem. MD5 jest rowniez zaszyfrowana w procesie.
//
//   Rozszyfrowanie polega na odtworzeniu (rozszyfrowaniu szumu) B
//   za pomoca blokow zaszyfrowanego hasla generatoem RSA. 
//   Szum binarny jest calkowice stochastyczny i generuje prawie
//   max lub max entropie Shannona. Podobnie jest z gneratorem RSA.
//   Stad do odszyfrowania konieczna jest znajomosc obydwu szumow
//   binarnych. Jesli plik zostanie zaszyfrowany ponownie to plik
//   wynikowy bedzie rozny w obydwu przypakach i wyliczona wartosc
//   entropi Shannona bedzie dla kaszdego z nich max lub prawie max
//   z niewielka fluktuacja probaibilistyczna. 
//        
//
//   Pliki wynikowe maja dlugosc ponad 2-krotna niz plik wyjsciowy.
//   Kazdy blok danych (max 4KB) ma wyliczona przed szyfrowaniem
//   skrot kontrolny, ktory jest sprawdzana ponownie po dekodowaniu.
//   Jesli wystapia bledy w odczytanych blokach to zostana zapisane
//   do plku o nazwie pliku wynikowego z rozszerzeniem ".err". 
//   Plik po szyfrowaniu pamieta nazwe pliku przed szyfroaniem tzn. 
//   ze jesli zostanie zmieniona nazwapliku szyfrowanego to po
//   odszyfrowaniu plik wynikowy bedzie mial pierwotna nazwe.
//
//   Do HEADER mozna dodac 3000 znakow opisu pliku. Aby to zrobic,
//   nalezy na sciezce, gdzie znajduje sie plik do szyfrowania, 
//   zapisac plik tekstowy z opisem (metada). Plik powinien miec
//   taka sama nazwe jak plik do szyfrowania (razem z postfix) 
//   z drugim postfixem ".met". Np. picture.jpeg dodajemy plik
//   tekstowy picture.jpeg.met (edytowany w dowolnym edytorze).
//   Plik tekstowy przy odczycie zachowuje znaki formatowania. 
//   Dodatkowy opis pliku ktory bedzie mial wiecej niz 3000 znakow
//   zostanie automatycznie obciety przez leszifr do 3000 znakow.
//
//   Pliki biblioteczne wymagane przy kompilacji za pomoca polecenia make:
//   gen_RSA.c           - generator RSA.
//   md5lib.c            - zmodyfikowana MD5 (dlugosc 32)
//   select_value.c      - operator rzutowy (w header uzywany jest U-FORMAT)
//   prandom_gen.c       - liniowy geneerator Gestego szumu binarnego
//   prandom_nonlinear.c - nieliniowy gen. Gestego szumu binarnego
//   ngen1.c             - coder/decoder oraz generator szumu binarnego FAST
//   Uwaga. pliki prandom oraz prandom_nonlinear sa wymienne i przed  
//   kompilacja nalezy zaznaczyc w sekcji #include z ktorym plikiem bedzie
//   kompilowany leszifr3.
//
//   Kompilacja: 
//       gcc -Wall leszifr3.c -o leszifr3 -mrdrnd
//   lub make    
//
///////////////////////////////////////////////////////////////////////////////


#define VERSION "Leszifr3 ver. 0.4 (c) for Linux 64-bit"


//dla ON/OFF zenityprogress bar w trybie cmd
int runlevelg = 0;


#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <time.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include "ngen1.c"		// ramdomizacja oraz genRSA (dla klucza)
#include "select_value.c"	// operator rzutowy dla q-string
#include "conv_chars_lib.c"	// converter/filter utf8 
#include "leszifr3.h"


//define TEST 0 - jest zdefiniowany w pliku md5lib.c


#define SHOW_ERROR 1 // 1 - pokazuje rodzaj bledu dla cmd mode, 0 - blokuje print


#define CODER 1   // 0 - test bez kodowania,   1 - koduje
#define DECODER 1 // 0 - test bez dekodowania, 1 - dekoduje


#define SFIX_FILEOUT ".archi"   //dodaje do nazwy pliku
#define SIZE_FILENAME 4096      //rozmiar path+file dla zenity

#define SPEED_CODING_LIMIT 3*1024*1024 //limit "slow coding" <-> "fast coding"

//dla detektora runlevel
int runlevel = 3; 





///////////////////////////////////////////////////////////////////////////////
// 				--- FUNKCJE STERUJACE ---
///////////////////////////////////////////////////////////////////////////////

//flaga stopujaca 100% progress_bar przy 100%
int runlevel3 = 1;

//runlevel 3 (progrss bar)
void progress_bar(int b){
   
    if(runlevel==-1) return;
 
    if(b>100 || !runlevel3) return;

    if(b<10) printf("\b\b\b\b\b\b\b\b\b\b\b\b");
    if(b>9 && b<100) printf("\b\b\b\b\b\b\b\b\b\b\b\b\b");
    if(b>99 && b<1000) printf("\b\b\b\b\b\b\b\b\b\b\b\b\b\b");
    fprintf(stdout, "progress: %d%s", b, "%");
    fflush(stdout);

    if(b==100 && runlevel3 )  {printf("\n"); runlevel3=0;}
}


//--- Graficzny progress bar ---
//otwiera progress bar dla runlevel5 (inicjuje graficzny progress bar
FILE*open_progress(void) {

   if(runlevel!=5) return NULL;

   char *s = "zenity --progress --percentage=0 --auto-close --text='coder/decoder is running ...' --no-cancel --width 400 2> /dev/null";

   FILE*pbar = popen(s, "w");
   return pbar;
}


//--- progress bar dla runlevel 5 ---
int runlevel5 = 0; //zbspiecza przed ponownym wjsciem po zmakieciu progress
//laduje dane do progress bar graficznego i zamyka po osiagnieciu 100%
void print_to_progress(FILE*f, int x) {

  if(runlevel!=5) return;
  if(runlevel5) return;

  if(x>=100) {
     //printf("%d\n", x);
     pclose(f);
     runlevel5 = 1;
     return;
  }

  fprintf(f, "%d\n", x);
  fflush(f);
}
//--- End graficzny progress bar
///////////////////////////////////////////////////////////////////////////////
//////////////////////////////// STERUJACE END ////////////////////////////////
///////////////////////////////////////////////////////////////////////////////




///////////////////////////////////////////////////////////////////////////////
// 			--- KODOWANIE/DEKODOWANIE ---
///////////////////////////////////////////////////////////////////////////////
//permutacja dla coder
int permut(char*buf, int size) {

   int i,j,k;
   int dv2 = size/2;

   char*per = (char*)malloc(size*sizeof(char));
   if(per==NULL) return -1;

   if(dv2>0) {
      for(i=0; i<dv2; i++) {
         j = 2*i;
         k=2*i+1;
         *(per+j) = *(buf+k);
         *(per+k) = *(buf+j);
      }
   }
   for(i=0; i<2*dv2; i++) *(buf+i) = *(per+i);

   free(per);
   return 0;
}


// --- coder (szyfrowanie) ---
int combine_streams(const char*fname, const char*key) {

   int i;
   int prog; //dla progress_bar;
   long long file_size;
   int blocks;
   int rest;
   int block_size;
   char*p;
   char s_file_size[512];
   char fname_out[512]="\0";

   char met[3000]; //metryczka moze miec do 3k znakow   
   int lenmet; //liczba znakow w metryczce

   //sumy kontrolne md5lib (dla bloku)
   char hash[40];
   unsigned int nhash;

   //flaga szybkosci kodowania
   //jesli filesize < SPEED_CODING_LIMIT to uruchamia "slow dense gen. noise"
   //jesli filesize >= SPEED_CODING_LIMIT to uruchamia "fast gen. noise" 
   int speed_coding = 0; 

   if(strlen(key) > 40) return -7; //klucz jest za dlugi
   if(strlen(key) < 4)  return -8; //lucz jest za krotki 

   FILE*fd;
   FILE*fw;
   FILE*pbar;

   struct stat sb;
   if(stat(fname, &sb) < 0) return -1; //brak pliku lub problem z jego open
   file_size = (long long)sb.st_size;
   if(file_size < 1L) return -2; //plik jest zerowy
   block_size = 4096; //dla zachowania zgodnosci z Windows

   //oblicza liczbe blokow 
   blocks = (int)(file_size / block_size);
   rest = (int)(file_size % block_size);

   //wybiera szybkosc generacji blokow szumu dla macierzy vizard
   if(file_size < SPEED_CODING_LIMIT) speed_coding = 1; 
   else speed_coding = 0;

   printf("Size:    %lld B\n", (long long) file_size); // rozmiar pliku
   printf("Sblock:  %d B\n", (int) block_size);        // rozmiar bloku max.4096B
   printf("Nblocks: %d\n", blocks);                    // liczba blokow
   printf("Rest:    %d B\n", rest);                    // reszta < rozmiar bloku

   //szybkosc kodowania
   if(speed_coding) 
   printf("Noise:   dense\n");
   else
   printf("Noise:   normal\n");

   //-----> tworzenie naglowka (tablicy head) <------
   int fname_len = (int)strlen(fname) + 10; //10 bytes na postfix 
   //zabezpieczenie przed przekroczniem rozmiaru tablicy head
   if(fname_len < block_size) fname_len = block_size; 
   char*head = (char*)malloc(fname_len*sizeof(char));
   if(head == NULL) return -3; //jesli nie mozna przyznac bloku pamieci

   //odczyt danych tekstowych metryczki z pliku. jesli metryczka istnieje
   lenmet = 0;
   *met = '\0';
   strcpy(head, fname);
   strcat(head, ".met");
   fd = fopen(head, "r");
   if(fd!=NULL) {
      while(fgets(head, 100, fd)!=NULL && lenmet < 3000) {
         lenmet += (int)strlen(head);
         if(lenmet < 3000) strcat(met, head);
      }
      fclose(fd);
   }   
   //---> Koniec odczytu danych z metryczki 

   printf("Meta:    %d chars\n", lenmet);
   
   //init head (naglowka) - naglowek nie powinen miec za duzo uporzadkowanej
   //wstepnie informacji (czyszczenie bufora operacyjnego)
   srand(time(NULL));
   for(i=0; i < 4096; i++) *(head+i) = rand() % 256; 

   //dodawnie do naglowka nazwy pliku
   strcpy(head, "FN=");
   p = strrchr(fname, '/');

   if(p!=NULL)  {
 
      //sprawdza czy dlugosc nazwy pliku nie przekracza zadeklarowanego
      //rzomairu bezpieczenstwa przekroczenia rozmiaru tablicy fname_out
      //(malo prawdopodobne) 
      if(strlen(p+1) > 512) {
           free(head);
           return -5;
      }
      strcat(head, p+1);

      //zabezpieczenie przeciw przekroczeniu zakresu dlugosci nazwy pliku max. 256
      //nazwa pliku +".archi" (SFIX_FILOUT) nie moze przekroczyc 256 znakow
      if((strlen(p+1) + strlen(SFIX_FILEOUT)) > 256) {
          free(head);
          return -5;
      }

      strcpy(fname_out, p+1);
      strcat(fname_out, SFIX_FILEOUT);

//printf("FNAME-OUT: %s\n", fname_out);
     
      //filtruje znaki diakrytyczne w utf-8
      filter_utf8_to_ascii(fname_out, (int)strlen(fname_out), s_file_size);
      strcpy(fname_out, s_file_size);

   } else { 

           //to samo jesli plik jestpobierany z tej samej sciezki co leszifr3
           if(strlen(fname) > 512) {
               free(head);
               return -5;
           }
           strcat(head, fname);

           //zabezpieczenie przeciw przekroczeniu zakresu max. 256
           //nazwa pliku do zapisu lacznie z ".archi" (SFIX_FILEOUT)
           if((strlen(fname) + strlen(SFIX_FILEOUT)) > 256) {
              free(head);
              return -5;
           }

           strcpy(fname_out, fname);
           strcat(fname_out, SFIX_FILEOUT);

           //filtruje znaki diakrytyczne w utf-8
           filter_utf8_to_ascii(fname_out, (int)strlen(fname_out), s_file_size);
           strcpy(fname_out, s_file_size);

          }

   //--- dodaje parametry pliku ---
   //q-string do header
   /*sprintf(s_file_size, "%lld", file_size);
   strcat(head, "&FILESIZE=");
   strcat(head, s_file_size);*/
   sprintf(s_file_size, "%d", block_size);
   strcat(head, "&BLOCKSIZE=");
   strcat(head, s_file_size);
   sprintf(s_file_size, "%d", blocks);
   strcat(head, "&NBLOCKS=");
   strcat(head, s_file_size);
   sprintf(s_file_size, "%d", rest);
   strcat(head, "&REST=");
   strcat(head, s_file_size);
   strcat(head, "&LENMET=");
   sprintf(s_file_size, "%d", lenmet);
   strcat(head, s_file_size);
   strcat(head, "&");

   //wstawia metryczke do head
   if(lenmet > 0) {
       memcpy(head+800, met, (size_t)lenmet + 1); //+1 dla '\0'
   }


   printf("\nSave: %s\n", fname_out);

   //----> otwiera plik do czytania <-----
   fd = fopen(fname,"r");
   if(fd==NULL) {free(head); return -1;} //brak pliku lub problem z jego open

   //nazwa pliku i otwieranie pliku do zapisu
   int szf =(int) strlen(fname) + (int)strlen(SFIX_FILEOUT);
   char*fname_write = (char*)malloc(szf*sizeof(char)+10);//<-- brak +10 smiecie w nazwie
   if(fname_write == NULL) {
       fclose(fd);
       free(head);
       return -3; //nie moze przydzelic pamieci
   }

   strcpy(fname_write, fname);
   strcat(fname_write, SFIX_FILEOUT);

   //---> otwiera plik do zapisu <-- Uwaga kasuje plik o nazwie pliku do zapisu
   fw = fopen(fname_write, "w");
   if(fw==NULL) {free(head); free(fname_write); fclose(fd); return -1;}

   //zwalnia nazwe pliku do zapisu po otwarciu deskryptora
   free(fname_write);

   //---> szyfrowanie
   gen_kmatrix(key); //tworzy macierz szyfrujaca
 

   //----> dla powtorzenia drugiej kopii zawartosci head <---- malloc
   char*head_rep = (char*)malloc(block_size*sizeof(char));
   if(head_rep == NULL) {
      free(head);
      fclose(fd);
      fclose(fw);
      return -3; //jesli nie mozna przyznac bloku pamieci
   }

   //kopiuje head -> head_rep
   memcpy(head_rep, head, block_size);

   //---> zapis pierwszej kopii naglowka
   //jesli rozmiar pliku jest ponizej limitu SPEED_CODING_LIMIT
   //to generowanie szumu przechodzi na tryb gesty szum
   if(speed_coding) {
      if(code_block(head, block_size) < 0) code_block_fast(head, block_size);
   }
   else code_block_fast(head, block_size); //tryb szybkiego gen. noise

   //permutacja wewnatrz bloku
   if(permut(head, block_size) < 0 ) {
      free(head);
      free(head_rep);
      fclose(fd);
      fclose(fw);
      return -3; //brak alokacji pamieci
   } 

   fwrite(vizard, (size_t)block_size, 1, fw);//zapis szyfrowanego szumu
   fflush(fw);
   if(ferror(fd)) {
      free(head);
      free(head_rep);
      fclose(fd);
      fclose(fw);
      return -4;
   }

   fwrite(head, (size_t)block_size, 1, fw); //zapis szyfrowanej tresci
   fflush(fw);
   if(ferror(fd)) {
      free(head);
      free(head_rep);
      fclose(fd);
      fclose(fw);
      return -4;
   }

   //---> zapis drugiej kopii naglowka
   //jesli rozmiar pliku jest ponizej limitu SPEED_CODING_LIMIT
   //to generowanie szumu przechodzi na tryb gesty szum
   if(speed_coding) {
      if(code_block(head_rep, block_size) < 0) code_block_fast(head_rep, block_size);
   }
   else code_block_fast(head_rep, block_size); //tryb szybkiego gen. noise

   //permutacja wewnatrz bloku
   if(permut(head, block_size) < 0 ) {
      free(head);
      free(head_rep);
      fclose(fd);
      fclose(fw);
      return -3; //brak alokacji pamieci
   } 

   fwrite(vizard, (size_t)block_size, 1, fw);//zapis szyfrowanego szumu
   fflush(fw);
   if(ferror(fd)) {
      free(head);
      free(head_rep);
      fclose(fd);
      fclose(fw);
      return -4;
   }

   fwrite(head_rep, (size_t)block_size, 1, fw); //zapis szyfrowanej tresci
   fflush(fw);
   if(ferror(fd)) {
      free(head);
      free(head_rep);
      fclose(fd);
      fclose(fw);
      return -4;
   }

   //---> zwalnia head_rep --- drugiej kopi naglowka
   free(head_rep);

   printf("\nStart coder:\n"); 

   int prog_pre = 1; //progrss bar zmienna predykcyjna

   // --- kodowanie blokow pliku ---
   if (blocks>0) {

      //Otwiera progress bar graficzny dla runlevel5
      if(runlevel==5) pbar = open_progress(); 
      prog = 0; //progress bar

      for(i=0; i<blocks; i++ ) {
         //odczyt
         fread(head, (size_t)block_size, 1, fd);
         if(ferror(fd)) {
            fclose(fd); 
            fclose(fw); 
            free(head);
	    if(prog < 100 && runlevel==5) pclose(pbar);
            return -4; 
         }

         //---> oblicza sume kontrona hash
         md5cnt(head, block_size, hash, &nhash);
         //---> tu szyfrowanie hash
         if(CODER) { 
             if(speed_coding){
                 if(code_block(hash, 40) < 0) code_block_fast(hash, 40);
             }
             else code_block_fast(hash, 40); //tryb szybkiego gen. noise
         }

   	 //permutacja wewnatrz bloku
         if(permut(hash, 40) < 0 ) {
             free(head);
             free(head_rep);
             fclose(fd);
             fclose(fw);
             return -3; //brak alokacji pamieci
         } 

        //---> zapis md5
        fwrite(vizard, 40, 1, fw); fflush(fw);
        if(ferror(fw)) {
           fclose(fd); 
           fclose(fw); 
           free(head);
           if(prog < 100 && runlevel==5) pclose(pbar);
           return -4; 
        }
        fwrite(hash, 40, 1, fw); fflush(fw);
        if(ferror(fw)) {
           fclose(fd); 
           fclose(fw); 
           free(head);
           if(prog < 100 && runlevel==5) pclose(pbar);
           return -4; 
        }
        //---> koniec zapisu md5

         //---> tu szyfrowanie bloku
         if(CODER) { 
             //szyfrowanie bloku szumem
             //jesli rozmiar pliku jest ponizej limitu SPEED_CODING_LIMIT
             //to generowanie szumu przechodzi na tryb gesty szum
             //jesli code_block zglosil blad to przechodzi na gen. fast noise
             if(speed_coding){
                 if(code_block(head, block_size) < 0) 
                    code_block_fast(head, block_size);
             }
             else code_block_fast(head, block_size); //tryb szybkiego gen. noise
         }

         //permutacja wewnatrz bloku
         if(permut(head, block_size) < 0 ) {
             free(head);
             free(head_rep);
             fclose(fd);
             fclose(fw);
             return -3; //brak alokacji pamieci
         } 

         //progress_bar (runlevel=3)
         prog = (i+1)*100/blocks;
         if(prog > prog_pre) {
            progress_bar(prog);
            prog_pre ++;
            //progress bar dla runlevel=5
            if(runlevel==5) print_to_progress(pbar, prog);
         } 
  
         //zapis zaszyfrowanego szumu
            fwrite(vizard, (size_t)block_size, 1, fw);
            fflush(fw);
            if(ferror(fw)) {
               fclose(fd); 
               fclose(fw); 
               free(head);
               if(prog < 100 && runlevel==5) pclose(pbar);
               return -4; 
            }
         
         //zapis zaszyfrowanej tresci
            fwrite(head, (size_t)block_size, 1, fw);
            fflush(fw);
            if(ferror(fw)) {
               fclose(fd); 
               fclose(fw); 
               free(head);
               if(prog < 100 && runlevel==5) pclose(pbar);
               return -4; 
            }
      } 
   }//if blocks

   // --- kodowanie reszty ---
   if(rest>0) {

      fread(head, (size_t)rest, 1, fd);
      if(ferror(fd)) {
          fclose(fd); 
          fclose(fw); 
          free(head);
          return -4; 
      }

      //---> oblicza sume kontrona hash
      md5cnt(head, rest, hash, &nhash);
      //---> tu szyfrowanie hash
      if(CODER) { 
         if(speed_coding){
             if(code_block(hash, 40) < 0) code_block_fast(hash, 40);
         }
         else code_block_fast(hash, 40); //tryb szybkiego gen. noise
      }

      //permutacja hash
      if(permut(hash, 40) < 0 ) {
         free(head);
         free(head_rep);
         fclose(fd);
         fclose(fw);
         return -3; //brak alokacji pamieci
      } 

      //---> zapis md5
      fwrite(vizard, 40, 1, fw); fflush(fw);
      if(ferror(fw)) {
         fclose(fd); 
         fclose(fw); 
         free(head);
         return -4; 
      }
      fwrite(hash, 40, 1, fw); fflush(fw);
      if(ferror(fw)) {
         fclose(fd); 
         fclose(fw); 
         free(head);
         return -4; 
      }
      //---> koniec zapisu sumy kontrolnej 

      //---> tu szyfrowanie reszty
      if(CODER) {
         //jesli rozmiar pliku jest ponizej limitu SPEED_CODING_LIMIT
         //to generowanie szumu przechodzi na tryb gesty szum
         if(speed_coding) {
             if(code_block(head, rest) < 0) code_block_fast(head, rest);
         }
         else code_block_fast(head, rest); //tryb szybkiego gen. noise
      }

      //permutacja reszty
      if(permut(head, rest) < 0 ) {
         free(head);
         free(head_rep);
         fclose(fd);
         fclose(fw);
         return -3; //brak alokacji pamieci
      } 

      //zapis zaszyfrowanego szumu   
         fwrite(vizard, (size_t)rest, 1, fw);
         fflush(fw);
         if(ferror(fw)) {
            fclose(fd); 
            fclose(fw); 
            free(head);
            return -4; 
         }

      //zapis zaszyfrowanej tresci   
         fwrite(head, (size_t)rest, 1, fw);
         fflush(fw);
         if(ferror(fw)) {
            fclose(fd); 
            fclose(fw); 
            free(head);
            return -4; 
         }
   }//if rest
   
   fclose(fd);
   fclose(fw);
   free(head);
   return 0;
}




//zapisuje bledy hash przy dekodowaniu
int write_hash_error_to_file(const char*fname, int nr_block) {
             
   int fname_len = (int)strlen(fname) + 10; //10 bytes na postfix
   char*fname_errors = (char*)malloc(fname_len*sizeof(char));
   if(fname_errors==NULL) return -1;

   strcpy(fname_errors, fname);
   strcat(fname_errors,".err");

   FILE*fd = fopen(fname_errors, "a");
   if(fd!=NULL) {
       //zapis numeru bloku gdzie byl wykryty blad
       fprintf(fd, "%d\n", nr_block); 
       fclose(fd);
   }
   return 0;
}



// --- decoder ---
//dekoduje plik zaszyfrowany przez combine_streams
int decode_streams(const char*fname, const char*key) {

   if(strlen(key) > 40) return -7; //klucz jest za dlugi
   if(strlen(key) < 4)  return -8; //lucz jest za krotki 

   struct stat sb;
   if(stat(fname, &sb)<0) return -1; //nie mozna otworzyc pliku
   long long file_size =(long long)sb.st_size;
   if(file_size<2*4096) return -1; //niepoprawny plik do otwarcia (zbyt krotki)

   //tworzy macierz deszyfrujaca dla zadanego klucza
   gen_kmatrix(key); //tworzy macierz szyfrujaca

   //dla sprawdzenia hash
   char hash_decoded[40];
   char hash[40];
   unsigned int nhash;
   //---> koniec hash

   char fname_out[512];
   char blocksize[40];
   char blocknumbers[80];
   char restlen[40];
  
   char buf[4096];
   char buf_rep[4096];
   char*p;
   int i, prog;

   FILE*pbar; //dla progress bar (zenity)

   //---> otwiera plik do odczyu <---
   FILE*fd  = fopen(fname, "r");
   if(fd==NULL) return -1;

   fread(vizard, 4096, 1, fd);
   if(ferror(fd)) {
      fclose(fd);
      return -1; //nie mozna odczytac poprawnie danych z pliku
   }
   fread(buf, 4096, 1, fd);
   if(ferror(fd)) {
      fclose(fd);
      return -1; //nie mozna odczytac poprawnie danych z pliku
   }

   //permutacja odwrotna pierwszej kopi naglowka
   if(permut(buf, 4096) < 0) {
       fclose(fd);
       return -3; //brak alokacji pamieci
   }

   //deszyfruje odczytany tekst (header) pierwszej kopi naglowka
   decode_block(buf, 4096); 
   
   //odczytuje druga kopie naglowka
   fread(vizard, 4096, 1, fd);
   if(ferror(fd)) {
      fclose(fd);
      return -1; //nie mozna odczytac poprawnie danych z pliku
   }
   fread(buf_rep, 4096, 1, fd);
   if(ferror(fd)) {
      fclose(fd);
      return -1; //nie mozna odczytac poprawnie danych z pliku
   }

   //permutacja odwrotna drugiej kopi naglowka
   if(permut(buf_rep, 4096) < 0) {
       fclose(fd);
       return -3; //brak alokacji pamieci
   }

   //deszyfruje odczytany tekst (header) drugiej kopi naglowka
   decode_block(buf_rep, 4096); 
   
   //---> selekcja danych z q-string za pomoca operatora rzutowego
   select_value(buf, "FN", fname_out);
   if(strcmp(fname_out, "NULL")==0) {
      //jesli nie mozna odczytac pierwszej kopii naglowka
      //to probuje zrobic to z druga kopia
      select_value(buf_rep, "FN", fname_out);
      if(strcmp(fname_out, "NULL")==0) {
          fclose(fd);
          return -2; //zle odczytany header
      } 
   }
   select_value(buf, "BLOCKSIZE", blocksize);
   if(strcmp(blocksize, "NULL")==0) {
      select_value(buf_rep, "BLOCKSIZE", blocksize);
      if(strcmp(blocksize, "NULL")==0) {
          fclose(fd);
          return -2;
      }
   }
   select_value(buf, "NBLOCKS", blocknumbers);
   if(strcmp(blocknumbers, "NULL")==0) {
      select_value(buf_rep, "NBLOCKS", blocknumbers);
      if(strcmp(blocknumbers, "NULL")==0) {
         fclose(fd);
         return -2;
      }
   }
   select_value(buf, "REST", restlen);
   if(strcmp(restlen, "NULL")==0) {
      select_value(buf_rep, "REST", restlen);
      if(strcmp(restlen, "NULL")==0) {
         fclose(fd);
         return -2;
      }
   }
   //---> Koniec dzialania operatora rzutowego

   //wyznacza parametry odczytu danych
   int block_size = atoi(blocksize);
   int blocks = atoi(blocknumbers);
   int rest = atoi(restlen);

   if(block_size>4096) {
      fclose(fd);
      return -3; //zla dlugosc bloku do odczytu
   }

   //tworzenie nazwy pliku do zapisu (malloc)
   int szf = (int)strlen(fname);
   szf += 512; //na nazwe pliku pobrana z headera 
   char*fname_write = (char*)malloc(szf*sizeof(char));
   if(fname_write==NULL) {
      fclose(fd);
      return -3; //brak alokacji pamieci
   }

   //obcina nazwe ".../<fname>.archi" i dodaje oryginalna nazwe pliku
   //jaka znajduje sie w header
   strcpy(fname_write, fname);
   p = strrchr(fname_write,'/');
   if(p!=NULL) {
      *(p+1) = '\0';
      strcat(fname_write, fname_out);
   } 
   else strcpy(fname_write, fname_out);

   //otwiera plik do zapisu
   FILE*fw = fopen(fname_write, "w");
   if(fw==NULL) {
      fclose(fd);
      free(fname_write);
      return -1;
   }

   int prog_pre = 1; //progrss bar zmienna predykcyjna

   //---> dekoduje bloki o rozmiarz block_size
   if(blocks>0) {
      //Otwiera progress bar graficzny dla runlevel5
      if(runlevel==5) pbar = open_progress();
      prog = 0; //progress bar

      for(i=0; i<blocks; i++ ) {

         //---> odczyt sumy kontrolnej
         fread(vizard, 40, 1, fd);
         if(ferror(fd)) {
             fclose(fd);
	     fclose(fw);
             remove(fname_write);
             if(prog < 100 && runlevel==5) pclose(pbar);
             free(fname_write);
             return -4;                     
         }
         fread(hash_decoded, 40, 1, fd);
         if(ferror(fd)) {
             fclose(fd);
	     fclose(fw);
             remove(fname_write);
             if(prog < 100 && runlevel==5) pclose(pbar);
             free(fname_write);
             return -4;                     
         }

         //permutacja odwrotna hash
	 if(permut(hash_decoded, 40) < 0) {
             fclose(fd);
	     fclose(fw);
             remove(fname_write);
             if(prog < 100 && runlevel==5) pclose(pbar);
             free(fname_write);
             return -3; 
         }

         if(DECODER)  decode_block(hash_decoded, 40); 
         //---> koniec sumy kontrolnej
  
         //---> odczyt bloku pliku
         fread(vizard, (size_t)block_size, 1, fd);
         if(ferror(fd)) {
             fclose(fd);
	     fclose(fw);
             remove(fname_write);
             if(prog < 100 && runlevel==5) pclose(pbar);
             free(fname_write);
             return -4;                     
         }
         fread(buf, (size_t)block_size, 1, fd);
         if(ferror(fd)) {
            fclose(fd);
	    fclose(fw);
            remove(fname_write);
            if(prog < 100 && runlevel==5) pclose(pbar);
            free(fname_write);
            return -4;                     
         }
         //---> koniec odczytu bloku danych

         //progress_bar
         prog = (i+1)*100/blocks;
         if(prog > prog_pre) {
              progress_bar(prog);
              prog_pre ++;
              //progress bar dla runlevel5
              if(runlevel==5) print_to_progress(pbar, prog);
         } 

         //---> deszyfruje odczytany tekst
         if(DECODER)  {
             if(permut(buf, block_size) <0) {
                 fclose(fd);
	         fclose(fw);
                 remove(fname_write);
                 if(prog < 100 && runlevel==5) pclose(pbar);
                 free(fname_write);
                 return -3; //brak alokacji pamieci
             }
             //dekoduje block
             decode_block(buf, block_size); 
             } 


         //---> obliczanie md5 dla bloku danych
         md5cnt(buf, block_size, hash, &nhash);
         if(strcmp(hash, hash_decoded)!=0) {
             //zapisuje numery blokow do pliku .err
             write_hash_error_to_file(fname_write, i);
         }     
         //---> koniec mapy bldow

         //---> zapis zdekodwanego bloku danych
         fwrite(buf, (size_t)block_size, 1, fw);
         fflush(fw);
         if(ferror(fw)) {
            fclose(fd);
	    fclose(fw);
            remove(fname_write);
            if(prog < 100 && runlevel==5) pclose(pbar);
            free(fname_write);
            return -4;
         }
      }
   }

   // ---> dekoduje reszte tzn. liczba danych mniejsza od bloku
   if(rest>0) {

      //---> odczyt sumy kontrolnej
      fread(vizard, 40, 1, fd);
      if(ferror(fd)) {
         fclose(fd);
	 fclose(fw);
         remove(fname_write);
         free(fname_write);
         return -4;                     
      }
      fread(hash_decoded, 40, 1, fd);
      if(ferror(fd)) {
         fclose(fd);
	 fclose(fw);
         remove(fname_write);
         free(fname_write);
         return -4;                     
      }

      //permutacja odwrotna hash
      if(permut(hash_decoded, 40) < 0) {
         fclose(fd);
	 fclose(fw);
         remove(fname_write);
         free(fname_write);
         return -3; 
      }

      if(DECODER)  decode_block(hash_decoded, 40); 
      //---> koniec hasza kontrolnego

      //---> odczyt reszty z pliku
      fread(vizard, (size_t)rest, 1, fd);
      if(ferror(fd)) {
         fclose(fd);
	 fclose(fw);
         remove(fname_write);
         free(fname_write);
         return -4;                     
      }
      fread(buf, (size_t)rest, 1, fd);
      if(ferror(fd)) {
         fclose(fd);
	 fclose(fw);
         remove(fname_write);
         free(fname_write);
         return -4;                     
      }
      //---> koniec odczytu reszty

      //deszyfruje odczytany tekst
      //if(DECODER) decode_block(buf, rest); 
      if(DECODER)  {
         if(permut(buf, rest) < 0) {
             fclose(fd);
	     fclose(fw);
             remove(fname_write);
             if(prog < 100 && runlevel==5) pclose(pbar);
             free(fname_write);
             return -3; //brak alokacji pamieci
         }
         //dekoduje rest
         decode_block(buf, rest); 
      } 

      //---> obliczanie md5 dla bloku danych
      md5cnt(buf, rest, hash, &nhash);
      if(strcmp(hash, hash_decoded)!=0) {
         //zapisuje numery blokow do pliku .err
         write_hash_error_to_file(fname_write, i+1);
      } 
      //---> koniec mapy bldow

      //zapis
      fwrite(buf, (size_t)rest, 1, fw);
      fflush(fw);
         if(ferror(fw)) {
            fclose(fd);
	    fclose(fw);
            remove(fname_write);
            free(fname_write);
            return -4;
         }
   }
  
   free(fname_write);
   fclose(fd);
   fclose(fw);
   return 0;
}


//odczytuje metryczke z headera
//zwraca zawrtosc metyczki w met (dklarowac char met[3000])
//zwraca:
//      0 - odczytal metryczke
//     -1 - brak pliku do odczytu lub nie moze otworzyc pliku
//     -2 - nie mozna odczytac naglowka
//     -3 - metryczka ma dlugosc zero tzn. nie byla dodana
int decode_met(const char*fname, const char*key, char*met) {

   if(strlen(key) > 40) return -7; //klucz jest za dlugi
   if(strlen(key) < 4)  return -8; //lucz jest za krotki 

   struct stat sb;
   if(stat(fname, &sb)<0) return -1; //nie mozna otworzyc pliku
   long long file_size =(long long)sb.st_size;
   if(file_size<2*4096) return -1; //niepoprawny plik do otwarcia

   //tworzy macierz deszyfrujaca dla zadanego klucza
   gen_kmatrix(key); //tworzy macierz szyfrujaca
  
   char buf[4096];
   char buf_rep[4096];
   char fname_out[512];
   char l_met[80];
 
   FILE*fd  = fopen(fname, "r");
   if(fd==NULL) return -1;

   fread(vizard, 4096, 1, fd);
   if(ferror(fd)) {
      fclose(fd);
      return -1; //nie mozna odczytac poprawnie danych z pliku
   }
   fread(buf, 4096, 1, fd);
   if(ferror(fd)) {
      fclose(fd);
      return -1; //nie mozna odczytac poprawnie danych z pliku
   }

   //permutacja odwrotna pierwszej kopi naglowka
   if(permut(buf, 4096) < 0) {
       fclose(fd);
       return -3; //brak alokacji pamieci
   }

   //deszyfruje odczytany tekst (header) - pierwsza kopia naglowka
   decode_block(buf, 4096); 


   fread(vizard, 4096, 1, fd);
   if(ferror(fd)) {
      fclose(fd);
      return -1; //nie mozna odczytac poprawnie danych z pliku
   }
   fread(buf_rep, 4096, 1, fd);
   if(ferror(fd)) {
      fclose(fd);
      return -1; //nie mozna odczytac poprawnie danych z pliku
   }

   //permutacja odwrotna drugiej kopi naglowka
   if(permut(buf_rep, 4096) < 0) {
       fclose(fd);
       return -3; //brak alokokacji pamieci
   }

   //deszyfruje odczytany tekst (header) drugiej kopi naglowka
   decode_block(buf_rep, 4096); 

   //---> zamyka plik do odczytu <----
   fclose(fd);
   
   //selekcja danych z q-string za pomoca operatora rzutowego
   select_value(buf, "FN", fname_out);
   if(strcmp(fname_out, "NULL")==0) {
       select_value(buf_rep, "FN", fname_out);
       if(strcmp(fname_out, "NULL")==0) return -2; //zle odczytany header
   }
   select_value(buf, "LENMET", l_met);
   if(strcmp(l_met, "NULL")==0) {
       select_value(buf_rep, "LENMET", l_met);
       if(strcmp(l_met, "NULL")==0) return -2; //zle odczytan header
   }
   
   if(atoi(l_met) == 0) {
      strcpy(met, "no added file description");
      return 0;
   } 

   memcpy(met, buf+800, (size_t)atoi(l_met)+1);
  
   return 0;
}
///////////////////////////////////////////////////////////////////////////////
//////////////////////// KODOWANIE/DEKODOWANIE END ////////////////////////////
///////////////////////////////////////////////////////////////////////////////




///////////////////////////////////////////////////////////////////////////////
//			--- PRINT ERRORS (CMD) ---
///////////////////////////////////////////////////////////////////////////////
//print errors
void print_error(int err) {

   puts(" ");
   if(err==-1)   puts(" Cannot read data from file [1]");
   if(err==-2)   puts(" Size of file is zero [2]");
   if(err==-3)   puts(" Cannot allocate memory [3]");
   if(err==-4)   puts(" Read or write on HDD is stopped [4]");
   if(err==-7)   puts(" Password too long max. 40 chars [7]");
   if(err==-8)   puts(" Password too short min. 4 chars [8]");
   if(err==-10)  puts(" Error of conversion string to number [10]");
   if(err==-11)  puts(" Incorrect comparison of redundant data [11]");
   if(err==-12)  puts(" Cannot decode the header [12]");
   if(err==-20)  puts(" Noise gen. doesn't work [20]");
   puts(" ");
}
///////////////////////////////////////////////////////////////////////////////
//			--- END PRINT ERRORS (CMD) ---
///////////////////////////////////////////////////////////////////////////////





///////////////////////////////////////////////////////////////////////////////
//				 --- ZENITY ---
///////////////////////////////////////////////////////////////////////////////
//sprawdza czy zenity jest zainstalowane w systemie
int check_zenity(void) {

   FILE*fp; 

   fp = popen("which zenity", "r");//dziala szybciej
   if(fp==NULL) return -1;
   char*buf = (char*)malloc(256*sizeof(char));
   if(buf==NULL) {pclose(fp); return -1;}
   fgets(buf, 256, fp);
   pclose(fp);
   char *p = strstr(buf, "zenity");
   if(p==NULL) {free(buf); return -1;}
   free(buf);
   return 0;
}


//opcja szyfrowanie/deszyfroanie
int zenity_select_option(void) {

   char buf[40];
   char cmd[20];
   char *p;
   char*s = "aw=$(zenity  --list  --text 'Select option' --radiolist  --column 'Pick' --column 'Option' TRUE Coder FALSE Decoder FALSE 'Read metadata' 2> /dev/null); echo $?; echo $aw";

   FILE*fp = popen(s, "r");
   if(fp==NULL) return -1;			//-1 - blad peopen
   fgets(cmd, 20, fp);
   p = strchr(cmd, '\n');
   if(p!=NULL) *p = '\0';
   if(strcmp(cmd, "1")==0) {pclose(fp); return -2;}
   fgets(buf, 20, fp);
   pclose(fp);
   p = strchr(buf, '\n');
   if(p!=NULL) *p = '\0';
   if(*buf=='\0') return -2;			// -2 - nie wprowadzono danych
   p = strstr(buf, "Decoder");
   if(p!=NULL) return 3;			//3 - Decoder
   p = strstr(buf, "Read metadata");	        //2 - Read metadata
   if(p!=NULL) return 2;
   return 1;					//1 - Coder
}


//otwieranie pliku (+path)
int zenity_frame_input_file(char*path_file) {

   char cmd[20];
   char *p;
   FILE*fp = popen("aw=$(zenity --file-selection 2> /dev/null); echo $?; echo $aw", "r");

   if(fp==NULL) return -1;
   fgets(cmd, 20, fp);
   p = strchr(cmd,'\n');
   if(p!=NULL) *p = '\0';
   if(strcmp(cmd, "1")==0) {pclose(fp); return -1;}
   fgets(path_file, 4096, fp);
   pclose(fp);
   p = strchr(path_file,'\n');
   if(p!=NULL) *p = '\0';
   if(*path_file=='\0') {strcpy(path_file, "null"); return -1;}
   return 0;
}


//wprowadzanie kluczy
//min dl. klucza = 4
//max dl. klucza = 40
//zzwraca 
//     1: klucze sa poprawnie wprowawdzone
//zwraca null/nul jesli blad
//    -1: problem z popen lub malloc lub brak kluczy
//    -2: (brak znaczenia tj. nie jest tu uzyte)
//    -3: klucz  za długi tj. więcej niż 40 znaków
//    -4: klucz jest za krotki
//    -5: klucz jest za dlugi
int zenity_input_keys(char*key) {

   strcpy(key, "null");

   char*s ="aw=$(zenity --forms --add-password='Key: ' --title='Password' --text='Enter the key' 2> /dev/null); echo $?; echo $aw";
   FILE*fp = popen(s, "r");
   if(fp==NULL) return -1;
   char reply[10];
   char*buf = (char*)malloc(SIZE_FILENAME*sizeof(char));
   if(buf==NULL) {pclose(fp); return -1;}
   
   //sprawdza czy bylo OK czy CANCEL
   fgets(reply, 10, fp);

   //sprawdza czy byl CANCEL
   char*p = strchr(buf, '\n');
   if(p!=NULL) *p = '\0';
   if(strcmp(reply, "1") == 0) {free(buf); pclose(fp); exit(0);}

   //pobiera druga linie z kluczem
   fgets(buf, 40, fp);
   pclose(fp);
   
   p = strchr(buf, '\n');
   if(p!=NULL) *p = '\0';

   //brak wpisanych kluczy lub zbyt krótki klucz min 4 chars
   if(strlen(buf) < 4 ) {free(buf); return -1;}
 
   if(strlen(buf)>40) {free(buf); return -3;} //too long key
   
   //separuje klucze
   strcpy(key, buf);
   free(buf);

   if(strlen(key)<3)  return -4; //key  jest za krotki
   if(strlen(key)>40) return -5; //key jest za dlugi

   return 1; 
}


//wyswietla ramke z errorerm
int zenity_show_error(const char*error) {

   char*s = "zenity --error --text="; 
   char*buf = (char*)malloc(SIZE_FILENAME*sizeof(char));
   if(buf==NULL) return -3; //brak alokacji pamieci

   strcpy(buf, s);
   strcat(buf, "'");
   strcat(buf, error);
   strcat(buf, "  ' 2> /dev/null");

   FILE*fp = popen(buf, "r");
   if(fp==NULL) {free(buf); return -1;}
   free(buf);
   pclose(fp);
   return 0;
}


//print errors
void zenity_codding_error(int err) {

   if(err==-1)   zenity_show_error(" Cannot read data from file or cannot open file [1]");
   if(err==-2)   zenity_show_error(" Size of file is zero or incorrect password [2]");
   if(err==-3)   zenity_show_error(" Cannot allocate memory [3]");
   if(err==-4)   zenity_show_error(" Read or write on HDD is stopped [4]");
   if(err==-7)   zenity_show_error(" Password too long max. 40 chars [7]");
   if(err==-8)   zenity_show_error(" Password too short min. 4 chars [8]");
   if(err==-10)  zenity_show_error(" Error of conversion string to number [10]");
   if(err==-11)  zenity_show_error(" Incorrect comparison of redundant data [11]");
   if(err==-12)  zenity_show_error(" Cannot decode the open file [12]");
   if(err==-20)  zenity_show_error(" Noise gen. does not work [20]");
}


//ramaka zbiorcza dla wprowadzania danych
int zenity(int ch_zenity) {

   int chk_zenity;
   if(ch_zenity==1) chk_zenity = 1; else chk_zenity = 0;

   int ret, r;
   char key1[50];

   //spradza czy zenity jest zainstalowane
   if(chk_zenity) {if(check_zenity()<0) return -1;} 

   ret = zenity_select_option();
   if(ret==-1) {
      zenity_show_error("Can't open a frame!");
      return -1;
   }
   if(ret==-2) return -1; //nie wprowadzono danych (user zrezygnowal)

   char*fname = (char*)malloc(SIZE_FILENAME*sizeof(char));
   if(fname == NULL) return -3; //brak alokacji pamieci 

   //---> kodowanie pliku
   if(ret==1) {

       //pobiera nazwe pliku (razem z path)
       r = zenity_frame_input_file(fname);
       if(r<0) {free(fname); return -1;}

       //wprawadza keys
       r =  zenity_input_keys(key1);
       switch(r) {
       case -1: zenity_show_error("No key or incorrect file");      return -1; break;
       case -3: zenity_show_error("Key is too long max. 40 chars"); return -1; break;
       case -4: zenity_show_error("Key is too short min. 4 chars"); return -1; break;
       case -5: zenity_show_error("Key is too long max. 40 chars"); return -1; break;
       case -6: zenity_show_error("Key is too short min.4 chars");  return -1; break;
       case -7: zenity_show_error("Key is too long max, 40 chars"); return -1; break; 
       //return -1; 
       }
   
       if(TEST) printf("fname: %s\n", fname);
       if(TEST) printf("key1:  %s\n", key1);
 
       //kodowanie 
       r = combine_streams(fname, key1);
       if(r<0) {
          zenity_codding_error(r);
          free(fname);
          return -1;
       }

       free(fname);
       return 0;
   }  


   //---> odczyt metryczki
   if(ret==2) {

       //pobiera nazwe pliku (razem z path)
       r = zenity_frame_input_file(fname);
       if(r<0) {free(fname); return -1;}

       //wprawadza keys
       r =  zenity_input_keys(key1);
       switch(r) {
       case -1: zenity_show_error("No key or incorrect file");      return -1; break;
       case -3: zenity_show_error("Key is too long max. 40 chars"); return -1; break;
       case -4: zenity_show_error("Key is too short min. 4 chars"); return -1; break;
       case -5: zenity_show_error("Key is too long max. 40 chars"); return -1; break;
       case -6: zenity_show_error("Key is too short min.4 chars");  return -1; break;
       case -7: zenity_show_error("Key is too long max, 40 chars"); return -1; break; 
       //return -1; 
       }
   
       if(TEST) printf("fname: %s\n", fname);
       if(TEST) printf("key1:  %s\n", key1);
 
       char*buf = (char*)malloc(3000*sizeof(char));
       if(buf==NULL) return -3; //brak alok8acji pamieci
       //kodowanie 
       r = decode_met(fname, key1, buf);
       if(r<0) {
          zenity_codding_error(r);
          free(fname);
          free(buf);
          return -1;
       }

       FILE*fd = fopen("tmpmet", "w");
       if(fd==NULL) {
          zenity_show_error("Internal problem with opening a file");
          free(fname);
          free(buf);
          return -1;
       }
       
       //pokazuje zawrtosc metryczki
       fprintf(fd, "%s\n", buf);
       fclose(fd);
       fd = popen("zenity --text-info --width 500 --height 400 --filename='tmpmet' 2>/dev/null", "r");
       if(fd == NULL) {
          zenity_show_error("Internal problem with opening a file");
          free(fname);
          free(buf);
          return -1;
       }
       pclose(fd);
       remove("tmpmet");

       free(buf);
       free(fname);
       return 0;
   }   


   //---> dekodowanie pliku
   if(ret==3) {

       //pobiera nazwe pliku (razem z path)
       r = zenity_frame_input_file(fname);
       if(r<0) {free(fname); return -1;}

       //wprowadza keys if(TEST) 
       r =  zenity_input_keys(key1);
       switch(r) {
       case -1: zenity_show_error("No key or incorrect file");        return -1; break;
       case -3: zenity_show_error("Keys are too long max. 40 chars"); return -1; break;
       case -4: zenity_show_error("Key1 is too short min. 3 chars");  return -1; break;
       case -5: zenity_show_error("Key1 is too long max. 40 chars");  return -1; break;
       //return -1;  
       }
   
       if(TEST) printf("fname: %s\n", fname);
       if(TEST) printf("key1:  %s\n", key1);

       //dekodowanie
       r = decode_streams(fname, key1);
       if(r<0) {
          zenity_codding_error(r);
          free(fname);
          return -1;
       }

       free(fname);
       return 0;

   }

   free(fname);
   zenity_show_error("No option selected");
   return 0;
}
// --- END GRAPHICS ZNITY ---


//---DETECTION RUN LEVEL ---
int detect_runlevel(void) {

   char buf[20];
   FILE*fp = popen("runlevel", "r");
   if(fp==NULL) return 3;
   fgets(buf, 20,fp);
   pclose(fp);
   char *p = strstr(buf, "5");
   if(p!=NULL) return 5;
   return 3;
}
///////////////////////////////////////////////////////////////////////////////
//////////////////////////////// ZENITY END ///////////////////////////////////
///////////////////////////////////////////////////////////////////////////////


//help
void print_help(void) {
   printf("\n");
   printf("Coding:    leszifr3 -c <file> <key>\n");
   printf("                       <file> - name of file to coding\n");
   printf("                       <key>  - required (use only ascii)\n\n");
   printf("Decode:    leszifr3 -d <file> <key>\n");
   printf("                       <file> - name of file to decoding\n");
   printf("                       <key>  - required key for decoding\n\n");
   printf("Metadata:  leszifr3 -m <file> <key>\n");
   printf("                       <file> - name of coded file\n");
   printf("                       <key>  - required key for decoding\n\n");
   printf("Version:   leszifr3 -v\n\n");
   printf("Gen. ver.: leszifr3 -vg\n\n");
   printf("\n");
}


//ON/OFF zenity progress bar w cmd
void switch_graphic_progress(int sw) {
      if(sw == 1) runlevelg = 1; else runlevelg = 0;
}


//szyfrowanie
int leszif3code(const char* fname, const char*passwd, int print) {
      int err;
      if(print==1) err = 1; else err = 0;
      if(err) runlevel =3; else runlevel = -1;
      if(runlevelg) runlevel = 5; //wlacza zenity progress bar
      int ret = combine_streams(fname, passwd);
      if(ret < 0 && err) print_error(ret);
      return ret;

}


//dekodowanie
int leszifr3decode(const char*fname, const char*passwd, int print) {
      int err;
      if(print==1) err = 1; else err = 0;
      if(err) runlevel = 3; else runlevel = -1;
      if(runlevelg) runlevel = 5; //wlacza zenity progress bar
      int ret = decode_streams(fname, passwd);
      if(ret < 0 && err) print_error(ret); 
      return ret;
}

//odczyt dodanego opisu pliku (dekaracja char out[3000])
//pobrany tekst w out zchowuje znaki sterujace
int leszifr3desc(const char*fname, const char*passwd, char*out, int print) {
      int err;
      if(print==1) err = 1; else err = 0;
      if(err) runlevel = 3; else runlevel = -1;
      int ret = decode_met(fname, passwd, out);
      if(ret<0 && err) print_error(ret); 
      return ret;
   }


//odczyt wersji pliku lub generatora
void leszifr3ver(const char*ver) {
    if(strcmp(ver, "-v")==0)  printf("\n %s\n\n", VERSION);
    if(strcmp(ver, "-vg")==0) {
        printf("\n");
        random_genenerator_version();
        printf("\n");
    }
}


//obsluga w modzie graficznym przez zenity
int run_with_zenity(int check_zenity) {
    runlevel = 5;
    int ret = zenity(check_zenity);
    return ret;
}


