
///////////////////////////////////////////////////////////////////////////////
// 				Generator szumu 
// 				    VIZARD
// 			   	    ver 0.3 
// 			   	    Library
// 			          (JPW 2018)
//
//    Generuje bloki vizard oraz kmatrix.
//    vizard - dla maskowania danych wejsciowych szumem
//    kmatrix - maska szyfru generowanego kluczem
//    
///////////////////////////////////////////////////////////////////////////////



#include <stdio.h>
#include <time.h>
#include <stdlib.h>
#include "md5lib.c"
#include "genRSA.c"
//#include "prandom_gen.c"      // generator liniowy
#include "prandom_nonlinear.c"  // generator nieliniowy



#define RAND_BLOCK_SIZE 4096 //MAX rozmiar bloku random numbers



//globalne maski szumu 
unsigned char vizard[RAND_BLOCK_SIZE];   //macierz szumu
unsigned char kmatrix[RAND_BLOCK_SIZE];  //mcierz generowana kluczem



///////////////////////////////////////////////////////////////////////////////
//                     --- generator szumu UNIX ---
///////////////////////////////////////////////////////////////////////////////
//dane generatora szumu
long long unix_seed;
long long m48;


//oblicza parametr m48
long long nu_m48(void) {
   int i;
   long long z = 1L;
   for(i=0; i<48; i++) z = z*2;
   return z-1; 
}


//ziarno
void srng48(long long s) {
   unix_seed = s;
   m48 = nu_m48();
}


//generatror rgen48 (UNIX)
long long rngen48(void) {
   long long a = 25214903917L;
   long long c = 11L;
   long long m = m48;
   unix_seed = (a*unix_seed + c) % m;
   return unix_seed;
}
///////////////////////////////////////////////////////////////////////////////
//                   --- END generatora szumu UNIX ---
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
//                        --- FAST NOISE GENERATOR ---
///////////////////////////////////////////////////////////////////////////////
// --- Niemonotoniczna funkcja czasu dla wyznaczania SEED ---
unsigned int non_monotonic_time_function(void) {

   struct timespec res;
   int ret;
   unsigned int nhash;

   char chrm[80];
   char hash[40];

   ret = clock_gettime(CLOCK_REALTIME, &res);

   if(ret) {
      sprintf(chrm, "%u", (unsigned int)time(NULL));       
      md5lib(chrm, hash, &nhash);
      srand(nhash);
      return (unsigned int) rand();
   }

   sprintf(chrm, "%ld", res.tv_sec);
   sprintf(hash, "%ld", res.tv_nsec);
   strcat(chrm, hash);
   md5lib(chrm, hash, &nhash);
   srand(nhash);
   return (unsigned int)rand();
}


//generator szumu unix (fast generator)
void noise_gen_unix(int no_zero) {

   int i,j,t;
   
   t = 4096/2;

   //parzyste
   unsigned int seed = non_monotonic_time_function();
   srng48((long long)seed);

   for(i=0; i < t; i++) {
      j = 2*i;
      if(no_zero == 1) {
         *(vizard+j) = 0;
         while(*(vizard+j) == 0)  *(vizard+j) = rngen48() % 256;
      } 
      else *(vizard+j) = rngen48() % 256;
   }

   //nieparzyste
   seed = non_monotonic_time_function();
   srng48((long long)seed);

   for(i=0; i < t; i++) {
      j = 2*i +1;
      if(no_zero == 1) {
         *(vizard+j) = 0;
         while(*(vizard+j) == 0)  *(vizard+j) = rngen48() % 256;
      } 
      else *(vizard+j) = rngen48() % 256;
   }
}
///////////////////////////////////////////////////////////////////////////////
//                        --- END FAST NOISE GENERATOR ---
///////////////////////////////////////////////////////////////////////////////




///////////////////////////////////////////////////////////////////////////////
//                        --- DENSE NOISE GENERATOR ---
///////////////////////////////////////////////////////////////////////////////
//generator szumu gesty
int noise_gen_dense(int lenbuf) {

   int ret;

   if(lenbuf > 4096) return -1;

   ret = random_block(vizard, lenbuf);
   if(ret < 0)  return -1;

   return 0;
}
///////////////////////////////////////////////////////////////////////////////
//                        --- END DENSE NOISE GENERATOR ---
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
//                        --- Szyfrowanie PASSWOD ---
///////////////////////////////////////////////////////////////////////////////
//produkuje matrix od klucza
void gen_kmatrix(const char*key) {

   int i;
   unsigned int nhash;
   char hash[40];

   md5lib(key, hash, &nhash);
   RSA_srand(nhash);

   for(i=0; i<RAND_BLOCK_SIZE; i++) 
        *(kmatrix+i) = (unsigned char)(RSA_rand() % 256);

}
///////////////////////////////////////////////////////////////////////////////
//                        --- END Szyfrowanie PASSWOD ---
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
//            --- Kodowanie blokow szumem gestym (slow coding) ---
///////////////////////////////////////////////////////////////////////////////
//szyfruje bloki 
//blok tekstu nie moze byc dluzszy niz 4kB
//najpierw nalezy uzyc gen_kamtrix aby otrzymac macierz szyfrujaca szum
//zwraca:
//      -1 - jesli nie mozna uruchomic generatora szumu 
//           lub text jest dluzszy niz RAND_BLOCK_SIZE
int code_block(char*text, int textlen) {

    int i, ret;

    //noise_gen(1); //generuje matrix vizard (noise)
    ret = noise_gen_dense(textlen); //inny mechanim tworzeniia szumu 
    if(ret < 0) return -1; //nie mozna uruchomic noise_gen dla vizard

    //szyfruje text z pomocą noise
    for(i=0; i<textlen; i++) *(text+i) = *(text+i) ^ *(vizard+i);

    //szyfruje noise za pomoca klucza
    for(i=0; i<textlen; i++) *(vizard+i) = *(vizard+i) ^ *(kmatrix+i);
    return 0;
}
///////////////////////////////////////////////////////////////////////////////
//                  --- Kodowanie blokow szumem gestym  ---
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
//              --- Kodowanie blokow szumem (fast coding)  ---
///////////////////////////////////////////////////////////////////////////////
//szyfruje bloki (fast noise generator)
//blok tekstu nie moze byc dluzszy niz 4kB
//najpierw nalezy uzyc gen_kamtrix aby otrzymac macierz szyfrujaca szum
void code_block_fast(char*text, int textlen) {
    int i;
    noise_gen_unix(1); //inny mechanim tworzeniia szumu (gen48. UNIX)
    //szyfruje text z pomocą noise
    for(i=0; i<textlen; i++) *(text+i) = *(text+i) ^ *(vizard+i);
    //szyfruje noise za pomoca klucza
    for(i=0; i<textlen; i++) *(vizard+i) = *(vizard+i) ^ *(kmatrix+i);
}
///////////////////////////////////////////////////////////////////////////////
//             --- END Kodowanie blokow szumem (fast coding)  ---
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
//                        --- De-kodowanie blokow   ---
///////////////////////////////////////////////////////////////////////////////
//deszyfruje bloki nie dłuzsze niz dlugosc tekstu
//najperw nalezy uzyc gen_kmatrix aby utowzyc matrix deszyfrujaca
void decode_block(char*text, int textlen) {

    int i;

    //deszyfruje noise za pomoca klucza
    for(i=0; i<textlen; i++) *(vizard+i) = *(vizard+i) ^ *(kmatrix+i);

    //deszyfruje text z pomocą noise
    for(i=0; i<textlen; i++) *(text+i) = *(text+i) ^ *(vizard+i);

}
///////////////////////////////////////////////////////////////////////////////
//                       --- END De-kodowanie blokow   ---
///////////////////////////////////////////////////////////////////////////////
