///////////////////////////////////////////////////////////////////////////////
//			Nieliniowy generator liczb losowych
//			            ver. 0.3
//                                 (JPW 2018)
//
//
//  Generator liczb losowych typu TRUE PSEUDO RANDOM.
//  Generator korzysta z dwoch zrodel entropi (SEED).
//  1. Generatora sprzetowyego Intel RDRAND (*)
//  2. Niemonotonicznej funkcji czasu (rodzielczosc nanosekundy)
//  (*) Intel wprowadzil funkcje RDRAND do CPU od III Generacji procesorow
//      i3/i5/i7 ("Ivy Bridge").
//      AMD dopiero w czerwcu 2015 roku. Z tego powodu w starszych PC
//      to zrodlo entropi (SEED) jest utomatycznie wylaczone. 
//
//  
//                             
//  kompilacja:
//             gcc -Wall prandom_nonlinear.c -mrdrnd
//             
///////////////////////////////////////////////////////////////////////////////



//wersja generatora
char*gen_ver = "0.3 (nonlinear) for Linux 64-bit";




#include <stdio.h>
#include <limits.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>
#include <cpuid.h>
#include <errno.h>


#define MAX_SEED 1000




///////////////////////////////////////////////////////////////////////////////
//                         --- PROTOYPY FUNKCJI ---
///////////////////////////////////////////////////////////////////////////////
//funkcja geneorwania bloku danych
//input:
//     buf     - bufor na probki random o wrtosciach (0...255)
//     lenbuf  - liczba probek tzn. romiar bufora buf
//zwraca:
//      -1 - nie udalo sie wygenerowac bloku random w buf
//       0 - blok zostal wygenerowany porawnie
int random_block(unsigned char*buf, int lenbuf);


//wersja generatora
void random_genenerator_version(void) {
     printf(" VERSION: %s\n", gen_ver);
}
///////////////////////////////////////////////////////////////////////////////
//                         --- END PROTOYPY FUNKCJI ---
///////////////////////////////////////////////////////////////////////////////




///////////////////////////////////////////////////////////////////////////////
//------------------- mnozenie i modulo duzych liczb integer ------------------
///////////////////////////////////////////////////////////////////////////////
//--- mnozy dwie bardzo dlugie liczby calkowite limit MAX_SEED ---
char*multiply_rnd(char a[],char b[]){

   static char mul[MAX_SEED];
   char c[MAX_SEED];
   char temp[MAX_SEED];
   int la,lb;
   int i,j,k=0,x=0,y;
   long int r=0;
   long sum = 0;

   la=strlen(a)-1;
   lb=strlen(b)-1;
   for(i=0;i<=la;i++) {
      a[i] = a[i] - 48;
   }
   for(i=0;i<=lb;i++) {
      b[i] = b[i] - 48;
   }

   for(i=lb;i>=0;i--) {
      r=0;
      for(j=la;j>=0;j--) {
         temp[k++] = (b[i]*a[j] + r)%10;
         r = (b[i]*a[j]+r)/10;
      }
      temp[k++] = r;
	 x++;
      for(y = 0;y<x;y++) {
   	 temp[k++] = 0;
      }
   }

   k=0;
   r=0;
   for(i=0;i<la+lb+2;i++) {
      sum =0;
      y=0;
      for(j=1;j<=lb+1;j++) {
  	 if(i <= la+j) {
	     sum = sum + temp[y+i];
 	 }
	 y += j + la + 1;
      }
      c[k++] = (sum+r) %10;
      r = (sum+r)/10;
   }

   c[k] = r;
   j=0;
   for(i=k-1;i>=0;i--) {
      mul[j++]=c[i] + 48;
   }

   mul[j]='\0';
   return mul;
}


// --- oblicza modulo dla very long  numbers limit (MAX_SEED) ---
unsigned long long mod_rnd(char*num, unsigned long long n) {
   if(n==0) return -1;
   int i;
   unsigned long long remainder = 0;
   for (i = 0; num[i] != '\0'; ++i)
      remainder = (remainder * 10 + num[i] - '0') % n;
   return remainder;
}
///////////////////////////////////////////////////////////////////////////////
//------------ END mnozenie i modulo duzych liczb calkowitych -----------------
///////////////////////////////////////////////////////////////////////////////




///////////////////////////////////////////////////////////////////////////////
//                    --- SOURCE OF ENTROPY RNRAND ---
/////////////////////////////////////////////////////////////////////////////// 
 /*
  funkcje uruchamiajace RDRAND (Intel/AMD)
  unsigned int __builtin_ia32_rdrand16_step (unsigned short *)
  unsigned int __builtin_ia32_rdrand32_step (unsigned int *)
  unsigned int __builtin_ia32_rdrand64_step (unsigned long long *)
*/


//sprawdza w cpu (intel/AMD) obecnosc funkcji RDRAND
//uwaga, funkcja RNRAND zostala wprowadzona do i3/i5/i7
//dopiero od III generacji "Ivy Bridge". 
//W przypadku AMD dopiero w czerwcu 2015.
int check_cpu(void) {
   const unsigned int flag_RDRAND = (1 << 30);
   unsigned int eax, ebx, ecx, edx;
  //nalezy ustawić rejestr eax=1 aby uzyskac info w trybie extended
  //zwraca status rdrand w ecx na pozycji 30 bitu 
   __get_cpuid(1, &eax, &ebx, &ecx, &edx);
   return (int)((ecx & flag_RDRAND) == flag_RDRAND);

} 


//gnerator ziarna -  pobiera z RDRAND (pierwsze zrodlo entropi) 
int randoms3(unsigned long long *rnd) {
    int retries= 10;
    unsigned long long rand64;

    while(retries--) {
        if ( __builtin_ia32_rdrand64_step(&rand64) ) {
            *rnd = rand64;
            return 1;
        }
    }
    return 0;
}


//--- zrodlo entropii RDRAND ---
//generuje seed w przypadku RNRAND (Intel/AMD)
//zwraca -1 jesli error
//zwraca 0 jesli wygenerowal seed
int seed_RDRAND(unsigned long long *seed) {

   //sprawdza obecnosc RNRAND w procesorze
   if(!check_cpu()) return -1;

   //tworzy seed_RNRAND
   if(!randoms3(seed)) return -1;

   return 0;
}
///////////////////////////////////////////////////////////////////////////////
//                         --- END SEED FROM RNRAND ---
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
// 				--- SEED FROM TIMER ---
///////////////////////////////////////////////////////////////////////////////
// --- zrodlo entropi (timer nasekundowy) ---
//niemonotoniczna funkcja czasu dokładnosc nansekundy
int nano_timer(unsigned long long *tms) {

   char sec[40], nsec[40];
   char stime[80];
   long long ns;
   int i;
   struct timespec res;

   if(clock_gettime(CLOCK_REALTIME, &res) < 0) return -1;

   sprintf(sec, "%ld", (long)res.tv_sec);
   sprintf(nsec, "%ld", res.tv_nsec);
   strcpy(stime, sec);
   strcat(stime, nsec);

   strcpy(stime, multiply_rnd(nsec, stime));

   ns = 1L;
   for(i=0; i<63; i++) ns = ns*2;
   ns = ns-1;

   *tms = mod_rnd(stime, ns);

   return 0;
}
///////////////////////////////////////////////////////////////////////////////
// 				--- SEED FROM TIMER ---
///////////////////////////////////////////////////////////////////////////////




///////////////////////////////////////////////////////////////////////////////
//                         --- FUNKCJE RANDOMIZE ---
///////////////////////////////////////////////////////////////////////////////
//tworzenie ziarna poczatkowego
int seed_creation(unsigned long long *seed1, unsigned long long *seed2) {
 
   int retA, retB;

   *seed1 = 0L; *seed2=0L;

   //z generatora entropi RDRND
   retA = seed_RDRAND(seed1);
 
   //retA = -1;//--> test if RDRND doesn't work

   //z generatora niemonotonicznej funkcji czasu
   retB = nano_timer(seed2);

   if(retA < 0) retA = nano_timer(seed1);

   int k = 0;
   while(*seed1==*seed2 && k < 7) {retA=nano_timer(seed2); k++;}

   if(retA < 0 || retB < 0) return -1;
   if(*seed1==*seed2) return -1;
   
   return 0;
}


unsigned long long xrnd1, xrnd2;
int init_rnd_true(void) {
     return seed_creation(&xrnd1, &xrnd2);
}


unsigned long long m62;
//--- oblicza parametr globalny:  m62 ---
void nu_m62(void) {
   int i;
   unsigned long long z = 1L;
   for(i=0; i<62; i++) z = z*2;
   m62 = z-1;
}

//nieliniowy generator true
unsigned long long prand_true(void) {
    
   char a[MAX_SEED], b[MAX_SEED], c[MAX_SEED];
   
   sprintf(a, "%llu", xrnd1);
   sprintf(b, "%llu", xrnd2);

   strcpy(c, multiply_rnd(a,b));
   xrnd2 = mod_rnd(c, m62);

   xrnd1 ^= xrnd2;

   return xrnd2;

}
///////////////////////////////////////////////////////////////////////////////
//                       --- END FUNKCJE RANDOMIZE ---
///////////////////////////////////////////////////////////////////////////////



///////////////////////////////////////////////////////////////////////////////
//                           --- block of random numbers ---
///////////////////////////////////////////////////////////////////////////////
//generuje dane randomize ze zbioru (0..255) do buf
//najlepiej uzyc buf o min. romizarze 4KB, raczej unikac maniejszych
//buforow na liczby random
int random_block(unsigned char*buf, int lenbuf) {

   int i,j;

   nu_m62();

   //inicjuje ziarno (1)
   if( init_rnd_true() < 0) return -1;

   int t= lenbuf/2;

   for(i=0; i<t; i++) {
      j = i*2; 
      *(buf+j) = (unsigned char)(prand_true() % 256);    
   }

   //inicjuje ziarno (2)
   if( init_rnd_true() < 0) return -1;

   for(i=0; i<t; i++) {
      j = i*2 + 1; 
      *(buf+j) = (unsigned char)(prand_true() % 256);    
   }

   return 0; 
}
///////////////////////////////////////////////////////////////////////////////
//                       ---  END block of random numbers ---
///////////////////////////////////////////////////////////////////////////////
